import os
import sys
import tempfile
import logging

from subprocess import Popen

import wpi.version


def set_logging():
    from wpi.log import set_stream_handler

    set_stream_handler()


def cur_file_dir():
    path = sys.path[0]
    if os.path.isdir(path):
        return path
    elif os.path.isfile(path):
        return os.path.dirname(path)

c_ = cur_file_dir()
p_ = os.path.realpath(os.path.join(cur_file_dir(), '..'))


def build(script, distpath, name=None, console=True, upx_dir=None, binarys=None,  output_specpath=None):
    import shutil
    binarys = binarys or []

    workpath = tempfile.mkdtemp()

    cmd = 'pyinstaller "{}" '.format(script)

    if name:
        cmd += '--name "{}" '.format(name)

    cmd += '--onefile '

    if output_specpath:
        cmd += '--specpath "{}" '.format(output_specpath)

    cmd += '--workpath "{}" '.format(workpath)
    cmd += '--distpath "{}" '.format(distpath)

    if upx_dir:
        cmd += '--upx-dir "{}" '.format(upx_dir)

    if console:
        cmd += '--console '
    else:
        cmd += '--windowed '

    cmd += '--uac-admin '

    for src, dest in binarys:
        cmd += '--add-binary "{}";"{}" '.format(src, dest)

    print('\npyinstaller cmd: \n', cmd, '\n')

    Popen(cmd, shell=True).wait()

    shutil.rmtree(workpath)


def run_verpatch(exe_path, verpatch_path):
    sys.path.append(p_)

    cmd = '{} {} ' \
          '{version} ' \
          '/va /pv {version} '\
          '/s description "Windows Printer Installer" '\
          '/s product "Windows Printer Installer" '\
          '/s copyright "Chen Meng, 2017" '\
          .format(verpatch_path, exe_path, version=wpi.version.__version__ + '.0', )

    print('\nverpathc cmd: \n', cmd, '\n')

    Popen(cmd, shell=True).wait()


def find_7z_path():
    regkeys = (r'SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall',
               r'SOFTWARE\Wow6432Node\Microsoft\Windows\CurrentVersion\Uninstall')
    from wpi.reg import Node

    for one in regkeys:
        n = Node(one)
        for k, subn in n.items():
            if k == '7-Zip':
                print(subn.tips['InstallLocation'])


def bundle_files():
    from wpi import ps_sample
    from wpi.env import get_szip_dir, SZIP_EXE, SZIP_DLL
    return (
        os.path.join(get_szip_dir(), SZIP_EXE),
        os.path.join(get_szip_dir(), SZIP_DLL),
        ps_sample.__file__,
    )


def main():
    import wpi.main

    from wpi import env
    from wpi import load_module

    from wpi2exe import config_

    os.chdir(tempfile.gettempdir())

    set_logging()

    config_dir = os.path.join(os.getenv('LOCALAPPDATA'), 'wpi2exe')
    os.makedirs(config_dir, exist_ok=True)

    config__filename = os.path.split(config_.__file__)[1]
    user_config__path = os.path.join(config_dir, config__filename)
    user_config_path = os.path.join(config_dir, 'config.py')

    wpi.main.copy_text_file(config_.__file__, user_config__path, even_exists=True)
    wpi.main.copy_text_file(config_.__file__, user_config_path, even_exists=False)

    config = load_module(user_config_path)

    verpatch_path = getattr(config, 'verpathc_path', None)
    upx_dir = getattr(config, 'upx_dir', None)

    if getattr(config, 'output_dir', None) is None:
        def_out_put = os.path.join(config_dir, 'exe')
        logging.warning('output_dir is None, use: {}'.format(def_out_put))
        output_dir = def_out_put
    else:
        output_dir = getattr(config, 'output_dir', None)

    output_filename = getattr(config, 'output_filename', None) or 'wpi'

    def do_build(console_, output_filename_):
        build(
            script=os.path.join(wpi.main.__file__),
            distpath=output_dir,
            name=output_filename_,
            console=console_,
            upx_dir=upx_dir,
            binarys=[(path, env.BUNDLE_DATA_FOLDER) for path in bundle_files() if path is not None]
        )

        if verpatch_path is not None:
            run_verpatch(verpatch_path=verpatch_path, exe_path=os.path.join(output_dir, output_filename_ + '.exe'))

    do_build(True, '{}-{}'.format(output_filename, wpi.version.__version__))
    # (do_build(False, '{}_nw-{}'.format(output_filename, wpi.version.__version__))


if __name__ == '__main__':
    main()



import io


def loads(strings: str):
    lines = [l.strip() for l in io.StringIO(strings).readlines()]
    lines = [l for l in lines if l != '']

    ls = []

    last_part = ''
    for line in lines:
        if line == '':
            continue

        if line.endswith('\\'):
            last_part += line
        else:
            ls.append(last_part+line)
            last_part = ''
            
    return _analyse(lines)


def _analyse(lines):
    data = {}

    sec = None

    for line in lines:
        if line[0] == ';':
            continue

        elif line[0] == '[':
            sec = _read_sec(line)
            data[sec] = []

        else:
            data[sec].append(_read_entry(line))
            
    return data


def _read_sec(line):
    start = line.index('[')
    end = line.rindex(']')

    v = line[start+1:end]
    return v


def _quote_handle(chars):
    v = ''
    in_quote = False

    i = 0
    for i in range(len(chars)):
        v += chars[i]

        if chars[i] == '"':
            if in_quote is True:
                _ = chars[i+1:].strip()
                if _ == '' or _[0] in (',', ';', '='):
                    break
            else:
                in_quote = True

    if v[0] != '"' or v[-1] != '"':
        raise Exception

    v = v[1:-1]
    while v != v.replace('""', '"'):
        v = v.replace()

    return v, chars[i+1:]


def _token_handle(chars):
    i = 0
    v = ''
    in_token = False

    for i in range(len(chars)):
        v += chars[i]

        if chars[i] == '%':
            if in_token is True:
                _ = chars[i+1:].strip()
                if _ == '' or _[0] in (',', ';', '='):
                    break
            else:
                in_token = True

    return v, chars[i+1:]


def _normal_handle(chars):
    i = 0
    v = ''

    for i in range(len(chars)):
        v += chars[i]

        _ = chars[i+1:].strip()
        if _ == '' or _[0] in (',', ';', '='):
            break

    return v, chars[i+1:]


def _read_entry(line):

    line = line.strip()

    v = None

    is_v_is_list = False
    v_list = []

    is_lr = False
    lv = None
    rv = None

    # is_lv_is_list = False
    lv_list = []

    is_rv_is_list = False
    rv_list = []

    while True:
        if line == '':
            break

        if line[0] == '"':
            quote_v, part_line = _quote_handle(line)
            line = part_line.strip()

            if is_lr:
                if is_rv_is_list:
                    rv_list[-1] += quote_v
                else:
                    rv = quote_v
            else:
                if is_v_is_list:
                    v_list[-1] += quote_v
                else:
                    v = quote_v

        elif line[0] == '%':
            token_v, part_line = _token_handle(line)
            line = part_line.strip()

            if is_lr:
                if is_rv_is_list:
                    rv_list[-1] += token_v
                else:
                    rv = token_v
            else:
                if is_v_is_list:
                    v_list[-1] += token_v
                else:
                    v = token_v

        elif line[0] == ';':
            break

        elif line[0] == ',':

            if is_lr:
                is_rv_is_list = True
                if len(rv_list) == 0:
                    rv_list = [rv, '']
                else:
                    rv_list.append('')
                rv = None

            else:
                is_v_is_list = True
                if len(v_list) == 0:
                    v_list = [v, '']
                else:
                    v_list.append('')
                v = None

            line = line[1:].strip()

        elif line[0] == '=':
            is_lr = True

            if is_v_is_list:
                lv_list = v_list
                del v_list

            else:
                lv = v
                v = None

            line = line[1:].strip()

        else:
            normal_v, part_line = _normal_handle(line)
            line = part_line.strip()

            if is_lr:
                if is_rv_is_list:
                    rv_list[-1] += normal_v
                else:
                    rv = normal_v
            else:
                if is_v_is_list:
                    v_list[-1] += normal_v
                else:
                    v = normal_v

    if is_lr:
        return (lv_list or lv), (rv_list or rv)

    elif is_v_is_list:
        return v_list

    else:
        return v



import logging


_is_file_handler_setted = False
_is_stream_handler_setted = False


def set_file_handler(filename):
    global _is_file_handler_setted
    if _is_file_handler_setted:
        return
    else:
        _is_file_handler_setted = True

    root_logger = logging.getLogger()
    root_logger.setLevel(logging.NOTSET)

    fh = logging.FileHandler(filename, 'w')
    fh_formatter = logging.Formatter(
        '[%(asctime)s %(filename)s fun:%(funcName)s line:%(lineno)d]: %(levelname)s: %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S')

    fh.setFormatter(fh_formatter)
    fh.setLevel(logging.NOTSET)
    root_logger.addHandler(fh)


def set_stream_handler():
    global _is_stream_handler_setted
    if _is_stream_handler_setted:
        return
    else:
        _is_stream_handler_setted = True

    root_logger = logging.getLogger()
    root_logger.setLevel(logging.NOTSET)

    ch = logging.StreamHandler()
    ch_formatter = logging.Formatter('  %(levelname)s: [%(filename)s:%(lineno)d]: %(message)s')
    ch.setFormatter(ch_formatter)
    ch.setLevel(logging.WARNING)
    root_logger.addHandler(ch)



import zipfile
import io
from subprocess import Popen, PIPE

import chardet


def is_can_handle(filename, path_of_7z=None):
    if zipfile.is_zipfile(filename):
        return True

    elif path_of_7z is not None:
        for bytes_line in io.BytesIO(
                Popen('"{}" t {}'.format(path_of_7z, filename), shell=True, stdout=PIPE).communicate()[0])\
                .readlines():

            if bytes_line.decode().strip() == 'Everything is Ok':
                return True

    return False


def list_names(archive, path_of_7z=None):

    if zipfile.is_zipfile(archive):
        with zipfile.ZipFile(archive) as z:
            return z.namelist()

    if path_of_7z is None:
        return

    list_head = '   Date      Time    Attr         Size   Compressed  Name'
    list_cut = '------------------- ----- ------------ ------------  ------------------------'

    _b = Popen('"{}" l -so {}'.format(path_of_7z, archive), shell=True, stdout=PIPE).communicate()[0]
    _code = chardet.detect(_b)['encoding']
    bytes_file = io.BytesIO(_b)
    lines = [line.decode(_code).rstrip() for line in bytes_file.readlines()]

    start_pos = None
    end_pos = None
    i = 0
    for line in lines:
        if line.startswith(list_head):
            start_pos = i + 2

        elif line.startswith(list_cut) and i > start_pos + 2:
            end_pos = i
            break

        i += 1

    return [line[53:] for line in lines[start_pos:end_pos]]


def read(archive, filename, path_of_7z=None):
    if zipfile.is_zipfile(archive):
        with zipfile.ZipFile(archive) as z:
            return z.read(filename)

    if path_of_7z is None:
        return None

    return Popen('"{}" e -so {} {}'.format(path_of_7z, archive, filename), shell=True, stdout=PIPE).communicate()[0]


def extract_all(archive, path, path_of_7z=None):
    if zipfile.is_zipfile(archive):
        with zipfile.ZipFile(archive) as z:
            z.extractall(path)
            return

    elif path_of_7z:
        Popen('"{}" x {} -o{}'.format(path_of_7z, archive, path), shell=True, stdout=PIPE).communicate()
    else:
        return None



from wpi.port import TCPIPPort, RAW, LPR, LocalPort


class ParameterError(Exception):
    pass


class RAWPort(TCPIPPort):
    def __init__(self, address, port=None, name=None, enable_snmp=False, snmp_dev_index=None, snmp_comunity=None):
        super().__init__()

        self.protocol = RAW
        self.address = address
        self.port = port or 9100

        self.__dict__['name'] = name or address + '_' + str(self.port)

        self.enable_snmp = enable_snmp
        self.snmp_dev_index = snmp_dev_index
        self.snmp_comunity = snmp_comunity


class LPRPort(TCPIPPort):
    def __init__(self, address, port=None, name=None,  enable_snmp=False, snmp_dev_index=None, snmp_comunity=None,
                 queue_name=None, is_enable_byte_count=False):
        super().__init__()

        self.protocol = LPR
        self.address = address
        self.port = port or 515

        self.__dict__['name'] = name or address + '_' + str(self.port)

        self.enable_snmp = enable_snmp
        self.snmp_dev_index = snmp_dev_index
        self.snmp_comunity = snmp_comunity

        self.queue_name = queue_name
        self.is_enable_byte_count = is_enable_byte_count


class SMBPort(LocalPort):
    def __init__(self, name):
        super().__init__(None)
        self.name = name


class Driver:
    def __init__(self, name, archive=None, inf_in_archive=None, inf_path=None):
        if bool(archive) and bool(inf_path):
            raise ParameterError

        self.name = name
        self.archive = archive
        self.inf_in_archive = inf_in_archive
        self.inf_path = inf_path


class Printer:
    def __init__(self, port, driver, name=None):
        self.port = port
        self.driver = driver
        self.name = name or driver


def ep(address, driver, name=None, protocol=None, ipport=None, archive=None, inf=None):

    if address.startswith('\\\\'):
        port = SMBPort(address)

    else:
        if protocol == RAW:
            port = RAWPort(address, ipport)
        elif protocol == LPR:
            port = LPRPort(address, ipport)

        elif protocol is None:

            if ipport is None:
                port = RAWPort(address)
            elif ipport >= 9100:
                port = RAWPort(address, ipport)
            else:
                port = LPRPort(address, ipport)

        else:
            raise ParameterError

    if archive:
        driver_obj = Driver(driver, archive, inf_in_archive=inf)
    else:
        driver_obj = Driver(driver, inf_path=inf)

    return Printer(port, driver_obj, name)



# Warning: Do not save data to this file when this file name is 'config_sample.py' and it under your home directory.
#          Cause it can be rewritten by program automatic.


# verpathc_path = r'Z:\verpatch.exe'


# upx_dir = r'Z:\\'


# output_dir = r'Z:\wpi'


# output_filename = 'wpi'





def is_token(string):
    try:
        if string[0] == '%' and string[-1] == '%':
            return True
        else:
            return False
    except IndexError:
        return False


def get_entries(infdata, section):
    for k, v in infdata.items():
        if k.lower() == section.lower():
            return v


def get_option(entries, option):
    for entry in entries:
        if entry[0].lower() == option.lower():
            return entry[1]


def get_models(infdata):
    old_models_heads = []
    for entry in get_entries(infdata, 'manufacturer'):
        v = entry[1]

        old_models_heads.append((v[0], None, None))

        for t in v[1:]:
            try:
                platform, osversion = t.split('.', 1)
            except ValueError:
                platform = t
                osversion = None

            old_models_heads.append((v[0], platform, osversion))

    new_models = {}
    for model in old_models_heads:
        sec = ''
        sec += model[0]
        if model[1] is not None:
            sec += '.' + model[1]
        if model[2] is not None:
            sec += '.' + model[2]

        entries = get_entries(infdata, sec)
        if entries is None:
            continue

        d2 = {}

        for entry in entries:
            d2[entry[0]] = entry[1][0], entry[1][1:]

        new_models[model] = d2

    return new_models


def get_provider(infdata):
    provider = get_option(get_entries(infdata, 'Version'), 'provider')
    if is_token(provider):
        provider = get_option(get_entries(infdata, 'Strings'), provider[1:-1])
    return provider


def get_driver_version(infdata):
    driverver = get_option(get_entries(infdata, 'version'), 'driverver')

    date = driverver[0]
    try:
        ver = driverver[1]
    except ValueError:
        ver = None
    return date, ver


def keep_platform(models, platform):
    platform = platform or (None, 'NTx86', 'NTamd64')
    new_models = {}
    for k, name_fh in models.items():
        if k[1] in platform:
            new_models[k] = name_fh

    return new_models


def clean(infdata, platform):
    models = get_models(infdata)

    models = keep_platform(models, platform)

    new_models = {}
    for k, name_fhs in models.items():
        new_name_fhs = {}
        for name, files_hids in name_fhs.items():
            new_name_fhs[name] = tuple(files_hids[1])

        new_models[k] = new_name_fhs

    return get_provider(infdata), get_driver_version(infdata), new_models



import logging

import hooky


class Printers(hooky.Dict):
    def __init__(self):

        from win32com.client import GetObject
        self._Printer = GetObject('winmgmts:/root/cimv2').Get('Win32_Printer')

        super().__init__()
        for _ in self._Printer.instances_():
            p = Printer(_)
            self.data[_.DeviceID] = p

    def save(self):
        for name, printer in self.data.items():
            printer.save(name)

    def make_new(self, name):
        if name in [_.lower() for _ in self.keys()]:
            raise KeyError

        win32obj = self._Printer.SpawnInstance_()

        new_printer = Printer(win32obj)

        self.data[name] = new_printer
        return new_printer

    def __delitem__(self, key):
        self.data[key].win32obj.Delete_()
        del self.data[key]


class Printer:
    def __init__(self, win32obj):
        self.win32obj = win32obj

        self._old_driver_name, self._old_port_name, self._old_name = self._load()

    def _load(self):
        return self.win32obj.DriverName, self.win32obj.PortName, self.win32obj.DeviceID

    def save(self, name):
        if self._old_driver_name != self.driver_name \
                or self._old_port_name != self.port_name \
                or self._old_name != name:

            self.win32obj.DeviceID = name
            logging.info('name: {}, driver name: {}, port name: {}'.format(name, self.driver_name, self.port_name))
            self.win32obj.Put_()

    @property
    def driver_name(self):
        return self.win32obj.DriverName

    @driver_name.setter
    def driver_name(self, v):
        self.win32obj.DriverName = v

    @property
    def port_name(self):
        return self.win32obj.PortName

    @port_name.setter
    def port_name(self, v):
        self.win32obj.PortName = v



import copy
import time
import logging

import win32serviceutil
from win32com.client import GetObject
from wpi import reg

import hooky

# import win32serviceutil

# Name -> 端口名

# HostAddress -> 主机名
# PortNumber -> 端口号, 一般为 9100/515 (lpr 强制)

# Queue -> 队列名称, 只限 lpr
# ByteCount -> 启用了 LPR 字节计数, 只限 lpr

# SNMPEnabled -> 启用了 SNMP 状态, True/False
# SNMPDevIndex -> SNMP 设备索引
# SNMPCommunity -> 社区名称, public（默认）/None,

# Protocol -> 协议 1(raw)/2(lpr)

# Put_() 保存
# Delete_() 删除

RAW = 1
LPR = 2

_LOCAL_PORT_REG_KEY = r'HKEY_LOCAL_MACHINE\Software\Microsoft\Windows NT\CurrentVersion\Ports'


class Ports(hooky.Dict):
    def __init__(self, computer=None, account=None):
        super(Ports, self).__init__()

        self._tcpipobj = GetObject('winmgmts:/root/cimv2').Get('Win32_TCPIPPrinterPort')
        self._localregtips = reg.Tips(_LOCAL_PORT_REG_KEY)
        self._old_localregtips = copy.deepcopy(self._localregtips)

        self._tcpipobjs_to_del = []

        self._tcpipobjs = {}
        for _ in self._tcpipobj.instances_():
            self.data[_.Name] = TCPIPPort(_)
            self._tcpipobjs[_.Name] = _

        for _, v in self._localregtips.items():
            self.data[_] = LocalPort(v)

    def make_a_tcpipport(self, name):
        if name in [_.lower() for _ in self.keys()]:
            raise KeyError

        win32obj = self._tcpipobj.SpawnInstance_()
        p = TCPIPPort(win32obj)
        self._tcpipobjs[name] = win32obj
        self[name] = p
        return p

    def make_a_localport(self, name):
        p = LocalPort()
        self[name] = p
        return p

    def save(self):
        for k, v in self.data.items():
            for one in self._tcpipobjs_to_del:
                one.Delete_()

            if isinstance(v, TCPIPPort):
                win32obj = self._tcpipobjs[k]
                win32obj.Name = k
                v.save_to(win32obj)
                win32obj.Put_()

            elif isinstance(v, LocalPort):
                self._localregtips[k] = v.value or ''

        if self._localregtips != self._old_localregtips:
            self._localregtips.save_to(_LOCAL_PORT_REG_KEY)
            self._old_localregtips = copy.deepcopy(self._localregtips)

            logging.info('restart print spooler')
            _ = 'Print Spooler'
            win32serviceutil.RestartService('Print Spooler')
            sec = 0
            while win32serviceutil.QueryServiceStatus(_)[1] != 4 and sec < 10:
                time.sleep(0.2)
                sec += 0.2

            print(sec)

    def __getitem__(self, key):
        for name, obj in self.data.items():
            if name.lower() == key.lower():
                return obj

    def __setitem__(self, key, value):
        if not isinstance(value, (TCPIPPort, LocalPort)):
            raise TypeError
        try:
            del self[key]
        except KeyError:
            pass

        try:
            del self._tcpipobjs[key]
        except KeyError:
            pass

        self.data[key] = value
        if isinstance(value, TCPIPPort):
            self._tcpipobjs[key] = self._tcpipobj.SpawnInstance_()

    def __delitem__(self, key):
        k_to_del = None
        for k, v in self.data.items():
            if k.lower() == key:
                k_to_del = key
                break

        if k_to_del is None:
            raise KeyError

        else:
            v = self.data[k_to_del]

            if isinstance(v, TCPIPPort):
                try:
                    self._tcpipobjs_to_del.append(self._tcpipobjs[key])
                    del self._tcpipobjs[key]
                except KeyError:
                    pass

            elif isinstance(v, LocalPort):
                del self._localregtips[key]

            del self.data[k_to_del]


class TCPIPPort:
    def __init__(self, win32obj=None):
        self.__dict__['_map'] = (
            ('protocol', 'Protocol', RAW),
            ('address', 'HostAddress', None),
            ('port', 'PortNumber', 9100),
            ('enable_snmp', 'SNMPEnabled', False),
            ('snmp_dev_index', 'SNMPDevIndex', 1),
            ('snmp_comunity', 'SNMPCommunity', 'public'),
            ('queue_name', 'Queue', '_'),
            ('is_enable_byte_count', 'ByteCount', False)
        )
        if win32obj:
            self.__dict__['data'] = self._load(win32obj)
        else:
            self.__dict__['data'] = {}
            for k, win32_k, default in self._map:
                setattr(self, k, default)

    def _load(self, win32obj):
        data = {}
        for k, win32_k, _ in self._map:
            data[k] = getattr(win32obj, win32_k)
        return data

    def save_to(self, win32obj):
        for k, win32_k, _ in self._map:
            setattr(win32obj, win32_k, getattr(self, k))

    def __getattr__(self, key):
        return self.data[key]

    def __setattr__(self, key, value):
        if key not in [one[0] for one in self._map]:
            raise AttributeError

        self.data[key] = value

    def __eq__(self, other):
        if self.protocol != other.protocol:
            return False
        if self.address != other.address or self.port != other.port:
            return False
        if self.enable_snmp != other.enable_snmp:
            return False
        if self.enable_snmp:
            if self.snmp_dev_index != other.snmp_dev_index or self.snmp_comunity != other.snmp_comunity:
                return False
        if self.protocol is LPR:
            if self.queue_name != other.queue_name or self.is_enable_byte_count != other.is_enable_byte_count:
                return False

        return True

    def update(self, other):
        if isinstance(other, TCPIPPort) or issubclass(type(other), TCPIPPort):
            self.data.update(other.data)
        else:
            self.data.update(other)


class LocalPort:
    def __init__(self, value=None):
        self.value = value



from wpi.des import ep, Printer, Driver, RAWPort, SMBPort, LPR

from wpi import add_credential

from wpi.env import CUR_BIT


# pa1 和 pa2 除了安装好后打印机列表里显示的名字不同，打印机端口和驱动都是一样的。
# ep 的意思是 easy printer， ep 其实是一个函数，返回一个 Printer 实例。
pa1 = ep('172.17.0.11', 'HP LaserJet 1022n', 'hp 1022n 1')

pa2 = Printer(
    port=RAWPort(address='172.17.0.11', port=9100, name='172.17.0.11_9100', enable_snmp=False),
    driver=Driver(name='HP LaserJet 1022n'),
    name='hp 1022n 2'
)


# pb1 和 pb2 也是除了显示的打印机名不一样，其它都一样。
# 现在使用的端口是 Windows 共享资源形式，姑且称之为 SMB 端口吧。
pb1 = ep(r'\\Printer-Server\HP LaserJet 1020', 'HP LaserJet 1020', 'hp 1020 1')

pb2 = Printer(
    port=SMBPort(r'\\Printer-Server\HP LaserJet 1020'),
    driver=Driver(name='HP LaserJet 1020'),
    name='hp 1020 2'
)

# 如果使用共享打印机需要提供账户密码才能使用，应该使用下面的函数添加凭据。
add_credential(host='Printer-Server', user='administrator', password='123456')


# 可以手动指定使用哪个包裹和在包裹里的哪个 inf 文件。
# archive 不是绝对路径时，程序寻找包裹时会自动拼接配置文件里定义的 drivers_dir。如果没有定义，则使用程序同目录下的 drivers 目录。
pc1 = ep('172.17.0.11', 'HP LaserJet 1022n', name='hp 1022n 3',
         archive='{bit}\lj1018_1020_1022-HB-pnp-win{bit}-sc.exe'.format(bit=CUR_BIT), inf='HPLJ1020.INF')


# 如果想使用以已不在包裹里的 inf 文件
pc2 = ep('172.17.0.11', 'HP LaserJet 1022n', name='hp 1022n 4',
         inf=r'{bit}\lj1018_1020_1022-HB-pnp-win{bit}-sc\HPLJ1020.INF'.format(bit=CUR_BIT))


# LPR 端口打印机
pd1 = ep('172.17.0.11', 'HP LaserJet 1022n', name='hp 1022n 5', protocol=LPR)


# 打印机集合文件里需要有一个名为 'printers' 的对象。
printers = [pa1, pa2, pb1, pb2, pc1, pc2, pd1]



import importlib.util
import os
from subprocess import Popen, PIPE

_modules = {}


def load_module(pypath, my_file=None):

    def _get_module_from_file_py35(_path):
        spec = importlib.util.spec_from_file_location(_path, _path)
        _module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(_module)
        return _module

    def _get_module_from_file_py34(_path):
        from importlib.machinery import SourceFileLoader
        _module = SourceFileLoader(_path, _path).load_module()
        return _module

    _p = ''

    if my_file is not None:
        _p = os.path.dirname(my_file)

    abs_path = os.path.abspath(os.path.join(_p, pypath))

    try:
        return _modules[abs_path]
    except KeyError:
        try:
            _modules[abs_path] = _get_module_from_file_py35(abs_path)
        except AttributeError:
            _modules[abs_path] = _get_module_from_file_py34(abs_path)

        return _modules[abs_path]


def add_credential(host, user, password):
    cmd = 'cmdkey /add:{} /user:{} /pass:{}'.format(host, user, password)
    p = Popen(cmd, shell=True, stdout=PIPE)
    p.wait()




from setuptools import setup
from distutils.util import convert_path

import ssl

ssl._create_default_https_context = ssl._create_unverified_context


NAME = 'wpi'

main_ns = {}
ver_path = convert_path('{}/version.py'.format(NAME))
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)

DESCRIPTION = 'Windows Printer Installer'


URL = 'https://github.com/meng89/' + NAME

CLASSIFIERS = [
    'Intended Audience :: Developers',
    'Intended Audience :: Information Technology',
    'Intended Audience :: System Administrators',
    'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
    'Operating System :: Microsoft :: Windows',
    'Programming Language :: Python :: 3',
    'Topic :: Software Development :: Libraries :: Python Modules',
]
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(name=NAME,
      version=main_ns['__version__'],
      description=DESCRIPTION,
      include_package_data=True,
      author='Chen Meng',
      author_email='ObserverChan@gmail.com',
      license='LGPL',
      url=URL,
      packages=[
          'wpi',
          'wpi.inf',
          'wpi2exe',
      ],
      entry_points={
          'console_scripts': [
               'wpi=wpi.main:main',
               'wpi2exe=wpi2exe.main:main'
          ],
      },
      install_requires=requirements,
      classifiers=CLASSIFIERS)



from win32com.client import GetObject
import hooky

_OBJ_NAME = 'Win32_PrinterDriver'

B32 = 'Windows NT x86'
B64 = 'Windows x64'


class Drivers(hooky.Dict):
    def __init__(self):
        super().__init__()
        self._Driver = GetObject('winmgmts:/root/cimv2').Get('Win32_PrinterDriver')

    def __iter__(self):
        for _ in self._Driver.instances_():
            yield tuple(_.name.rsplit(',', 2))

    def __setitem__(self, key, value):
        raise TypeError('not support this yet!')

    def add_by_inf(self, inf_path, name, platform=None):
        if inf_path is None:
            raise ValueError

        self._Driver.Name = name
        self._Driver.InfName = inf_path

        # default is os platform
        if platform is not None:
            self._Driver.SupportedPlatform = platform

        method = self._Driver.Methods_('AddPrinterDriver')
        in_parms = method.InParameters
        in_parms.DriverInfo = self._Driver

        self._Driver.ExecMethod_('AddPrinterDriver', in_parms)

    def __delitem__(self, key):

        wmi = GetObject('winmgmts:/root/cimv2')

        drivers = wmi.InstancesOf(_OBJ_NAME)

        for driver in drivers:
            if key == driver.name.rsplit(',', 2):
                driver.Delete_()
                return

        raise Exception


if __name__ == '__main__':
    pass



import collections

import winreg

from hooky import Dict


ERROR_NO_MORE_ITEMS = 259


def _recommend_type(v):
    if isinstance(v, int):
        if 0 <= v <= 2**32-1:
            return winreg.REG_DWORD
        elif 2**32 <= v <= 2**64-1:
            return winreg.REG_QWORD
        else:
            return None

    elif isinstance(v, str):
        if v.startswith('%') and v.endswith('%'):
            return winreg.REG_EXPAND_SZ
        else:
            return winreg.REG_SZ

    elif isinstance(v, collections.Iterable):
        return winreg.REG_MULTI_SZ


def _analyze_path(path):
    if isinstance(path, (tuple, list)):
        return tuple(path)

    else:
        if isinstance(path, str):
            return tuple(path.split('\\'))
        else:
            raise Exception


class Node(Dict):
    def __init__(self, path=None, skip_error=True):
        super().__init__()

        if path:
            self.tips, self.data = self._load(path, skip_error)

        else:
            self.tips = Tips()

    @staticmethod
    def _load(path, skip_error=True):
        path = _analyze_path(path)

        tips = Tips(path, skip_error)

        subs = {}

        try:
            my_key = winreg.OpenKey(getattr(winreg, path[0]),
                                    '\\'.join(path[1:]),
                                    0,
                                    winreg.KEY_ALL_ACCESS | winreg.KEY_WOW64_64KEY)
        except Exception as e:
            if skip_error:
                return tips, subs
            else:
                raise e

        i = 0
        while True:
            try:
                k = winreg.EnumKey(my_key, i)
            except Exception as E:
                if isinstance(E, OSError) and E.winerror == ERROR_NO_MORE_ITEMS:
                    break
                elif skip_error:
                    continue
                else:
                    raise E

            subs[k] = Node(path + (k,), skip_error)
            i += 1

        my_key.Close()

        return tips, subs

    def save_to(self, path, skip_error=True):
        path = _analyze_path(path)

        try:
            my_key = winreg.CreateKeyEx(getattr(winreg, path[0]),
                                        '\\'.join(path[1:]),
                                        0,
                                        winreg.KEY_ALL_ACCESS | winreg.KEY_WOW64_64KEY)
        except Exception as E:
            if skip_error:
                return
            else:
                raise E

        # Tips
        self.tips.save_to(path, skip_error)

        # Subs
        reg_keys = set()

        i = 0
        while True:
            try:
                sub_key = winreg.EnumKey(my_key, i)
            except Exception as E:
                if isinstance(E, OSError) and E.winerror == ERROR_NO_MORE_ITEMS:
                    break
                elif skip_error:
                    i += 1
                    continue
                else:
                    raise E

            i += 1
            reg_keys.add(sub_key)

        # keys_to_add = set(self.keys()) - reg_keys
        keys_to_del = reg_keys - set(self.keys())

        for k in keys_to_del:
            winreg.DeleteKey(my_key, k)

        my_key.Close()

        # save Subs
        for k, v in self.items():
            v.save_to(path+(k,))

    def __setitem__(self, key, value):
        if not isinstance(value, Node):
            raise TypeError

        k_to_del = None
        for k in self.keys():
            if k.lower() == key.lower():
                k_to_del = k
                break
        if k_to_del:
            del self[k_to_del]

        self.data[key] = value


class Tips(Dict):
    def __init__(self, path=None, skip_error=True):
        super().__init__()

        if path:
            self.data, self.types = self._load(path, skip_error)
        else:
            self.types = {}

    @staticmethod
    def _load(path, skip_error=True):
        path = _analyze_path(path)

        data = {}
        types = {}

        try:
            my_key = winreg.OpenKey(getattr(winreg, path[0]),
                                    '\\'.join(path[1:]),
                                    0,
                                    winreg.KEY_ALL_ACCESS | winreg.KEY_WOW64_64KEY)
        except Exception as E:
            if skip_error:
                return data, types
            else:
                raise E

        i = 0
        while True:
            try:
                name, value, type_ = winreg.EnumValue(my_key, i)
            except Exception as E:
                if isinstance(E, OSError) and E.winerror == ERROR_NO_MORE_ITEMS:
                    break
                elif skip_error:
                    i += 1
                    continue
                else:
                    raise E
            i += 1
            data[name] = value
            types[name] = type_

        my_key.Close()

        return data, types

    def save_to(self, path, skip_error=True):
        path = _analyze_path(path)
        try:
            my_key = winreg.CreateKeyEx(getattr(winreg, path[0]),
                                        '\\'.join(path[1:]),
                                        0,
                                        winreg.KEY_ALL_ACCESS | winreg.KEY_WOW64_64KEY)
        except Exception as e:
            if skip_error:
                return
            else:
                raise e

        reg_data, reg_types = self._load(path)

        keys_to_del = set(reg_data.keys()) - set(self.keys())
        keys_to_add = set(self.keys()) - set(reg_data.keys())

        keys_to_update = set([k for k in (set(self.keys()) & set(reg_data.keys())) if self[k] != reg_data[k]])

        # Add and Update
        for k in (keys_to_add | keys_to_update):
            v, type_ = self.get_value(k)
            try:
                winreg.SetValueEx(my_key, k, 0, type_, v)
            except Exception as e:
                if skip_error:
                    return
                else:
                    raise e
        # Del
        for k in keys_to_del:
            try:
                winreg.DeleteValue(my_key, k)
            except Exception as e:
                if skip_error:
                    return
                else:
                    raise e

        my_key.Close()

    def get_value(self, key):
        try:
            _type = self.types[key]
        except KeyError:
            _type = _recommend_type(self[key])
        return self[key], _type

    def set_value(self, key, type_, value):
        try:
            del self[key]
        except KeyError:
            pass
        try:
            del self.types[key]
        except KeyError:
            pass

        self[key] = value
        self.types[key] = type_



import logging
import os
import platform
import sys

BUNDLE_DATA_FOLDER = '_data'

SZIP_EXE = '7z.exe'
SZIP_DLL = '7z.dll'


def is_exe():
    if getattr(sys, 'frozen', False) is not False:
        return True
    else:
        return False


def meipass_path():
    return getattr(sys, '_MEIPASS')


def exe_path():
    if is_exe():
        return sys.executable
    elif __file__:
        return __file__


def exe_dir():
    return os.path.dirname(exe_path())


def _find_7z_in_reg():
    from wpi.reg import Node

    b32_location = None
    b64_location = None

    n = Node(r'HKEY_LOCAL_MACHINE\SOFTWARE\7-Zip')
    for k, v in n.tips.items():
        if k == 'Path32':
            b32_location = v
        elif k == 'Path64':
            b64_location = v

    return b32_location, b64_location


def get_szip_dir():
    if is_exe():
        z7_dir = os.path.join(meipass_path(), BUNDLE_DATA_FOLDER)
        if os.path.isfile(z7_dir):
            return z7_dir

    szip_32_dir, szip_64_dir = _find_7z_in_reg()

    if szip_32_dir and os.path.isfile(os.path.join(szip_32_dir, SZIP_EXE)):
        return szip_32_dir

    elif szip_64_dir and os.path.isfile(os.path.join(szip_64_dir, SZIP_EXE)):
        return szip_64_dir

    else:
        logging.warning('7-Zip cannot be found.')
        return None


def get_szip_path():
    return os.path.join(get_szip_dir(), SZIP_EXE)


class Config:
    __slots__ = ['z7_path', 'drivers_dir', 'archive_exts']

    def __init__(self, obj=None):
        for k in self.__slots__:
            setattr(self, k, getattr(obj, k, None))


def load_config(path):
    from wpi import load_module

    return Config(load_module(path))


def supply_config(config=None):
    c = Config(config)

    c.z7_path = c.z7_path or get_szip_path()

    c.archive_exts = c.archive_exts or ['.zip', '.7z', '.rar', '.exe']

    return c


ARCHIVE_EXTS = ['.zip', '.7z', '.rar', '.exe']


CUR_OS = platform.release().lower()
ALL_OS = {'xp', '7', '10'} | {CUR_OS}


B32 = '32'
B64 = '64'
ALL_BITS = {'32', '64'}


if platform.machine().endswith('64'):
    CUR_BIT = '64'
elif platform.machine().endswith('86'):
    CUR_BIT = '32'
else:
    raise Exception


PYTHON_BIT = platform.architecture()[0][0:2]



import ctypes
import datetime
import logging
import os
import sys
import tempfile

import chardet

import wpi.inf
import wpi.log
from wpi import load_module, archivelib, version
from wpi.env import ALL_BITS, CUR_BIT, CUR_OS, ALL_OS, PYTHON_BIT

DEFAULT_PS_NAME = 'ps.py'

USER_SAMPLE_PS_NAME = '_ps.py'

DEFAULT_DIRIVERS_NAME = 'drivers'

LOGS_DIR = os.path.join(os.getenv('LOCALAPPDATA'), 'wpi_logs')


def split_all(path):
    names = []
    while True:
        path, name = os.path.split(path)

        if name != '':
            names.append(name)
        else:
            if path != '':
                names.append(path)
            break

    names.reverse()
    return names


def check_dirs_names(names: set, must_have: set, must_have_no: set, case_sensitive=False):
    if case_sensitive:
        if not names & must_have_no:
            if names >= must_have:
                return True
    else:
        if not (set([_.lower() for _ in names]) & set([_.lower() for _ in must_have_no])):
            if set([_.lower() for _ in names]) >= set([_.lower() for _ in must_have]):
                return True

    return False


def list_files(d):
    all_files = []

    for root, dirs, files in os.walk(d):
        for file in files:
            all_files.append(os.path.join(root, file))

    return all_files


def _get_models(inf_bytes):

    try:
        inf_data = wpi.inf.loads.loads(inf_bytes.decode(chardet.detect(inf_bytes)['encoding']))
        models = wpi.inf.utils.get_models(inf_data)
    except Exception:
        return {}
    else:
        return models


def list_infs(archive):
    from wpi.archivelib import list_names
    from wpi.env import get_szip_path

    return [_ for _ in list_names(archive, get_szip_path()) if _.lower().endswith('.inf')]


def is_match(inf_bytes, driver):
    models = _get_models(inf_bytes)

    for model, namek_files_hids in models.items():
        if CUR_BIT == '32' and model[1] is not None and model[1].lower() != 'ntx86':
            continue

        elif CUR_BIT == '64' and (model[1] is None or model[1].lower() != 'ntamd64'):
            continue

        if driver in namek_files_hids.keys():
            return True

    return False


def get_legal_infs_list(archive, driver, z7_path):
    legal_infs = []

    for inf in list_infs(archive):
        inf_bytes = archivelib.read(archive, inf, z7_path)
        if is_match(inf_bytes, driver):
            legal_infs.append(inf)

    return legal_infs


def filter_files_by_exts(files, exts, case_sensitive=False):
    new_files = []
    if case_sensitive:
        for file in files:
            if os.path.splitext(file)[1] in exts:
                new_files.append(file)
    else:
        for file in files:
            if os.path.splitext(file)[1].lower() in [ext.lower() for ext in exts]:
                new_files.append(file)
    return new_files


def filter_files_by_isarchive(files, z7_path):
    files = [file for file in files if archivelib.is_can_handle(file, z7_path)]
    return files


def filter_files_by_dirs(files, must_have, must_have_no, drivers_dir):
    new_files = []

    delete_root = drivers_dir

    for path in files:
        if not path.startswith(delete_root):
            raise Exception

        _part_path = path[len(delete_root):]
        part_path = os.path.split(_part_path)[0]

        if check_dirs_names(set(split_all(part_path)), must_have=must_have, must_have_no=must_have_no):
            new_files.append(path)

    return new_files


def _get_archive_infs_list(driver, must_have, drivers_dir, archive_exts, z7_path):

    archives = filter_files_by_isarchive(
                filter_files_by_dirs(
                    filter_files_by_exts(list_files(drivers_dir), archive_exts),
                    must_have,
                    (ALL_BITS | ALL_OS) - must_have,
                    drivers_dir
                ),
                z7_path
            )

    legal_archive_infs_list = []

    for archive in archives:
        legal_infs = get_legal_infs_list(archive, driver, z7_path)

        if legal_infs:
            legal_archive_infs_list.append([archive, legal_infs])

    return legal_archive_infs_list


class InstallationFailed(Exception):
    pass


def install_port(des_port):
    from wpi.port import Ports

    from wpi.des import RAWPort, LPRPort, SMBPort

    sysports = Ports()

    if des_port.name is None:
        raise Exception

    if des_port.name not in sysports.keys():
        if isinstance(des_port, (RAWPort, LPRPort)):
            sysports[des_port.name] = des_port

        elif isinstance(des_port, SMBPort):
            sysports[des_port.name] = des_port

        sysports.save()
    else:
        logging.warning('port: {} already existed, abort.'.format(repr(des_port.name)))


def install_driver(des_driver, drivers_dir):
    from wpi.driver import Drivers, B32, B64
    from wpi import env
    from wpi.env import ARCHIVE_EXTS, get_szip_path

    szip_path = get_szip_path()

    sysdrivers = Drivers()

    for one in sysdrivers:
        if one[0] == des_driver.name and (one[2], CUR_BIT) in ((B32, env.B32), (B64, env.B64)):
            logging.warning('driver: {} already existed, abort.'.format(repr(des_driver.name)))
            return None

    inf_path = None

    tempdir = tempfile.mkdtemp()

    def _full(_):
        if not os.path.isabs(_):
            return os.path.join(drivers_dir, _)
        return _

    if des_driver.inf_path:
        inf_path = _full(des_driver.inf_path)

    elif des_driver.archive:

        archive = _full(des_driver.archive)

        if des_driver.inf_in_archive:
            iia = des_driver.inf_in_archive
        else:
            infs = get_legal_infs_list(archive, des_driver.name, szip_path)
            iia = infs[0]

        logging.info('use archive: {}'.format(repr(archive)))

        archivelib.extract_all(archive, tempdir, szip_path)

        inf_path = os.path.join(tempdir, iia)

    else:
        for must_have in ({CUR_BIT, CUR_OS}, {CUR_BIT}):
            archive_infs = _get_archive_infs_list(des_driver.name, must_have, drivers_dir, ARCHIVE_EXTS, szip_path)

            if archive_infs:
                archive = archive_infs[0][0]
                logging.info('use matched archive: {}'.format(repr(archive)))
                iia = archive_infs[0][1][0]
                logging.info('inf_in_archive: {}'.format(repr(iia)))

                archivelib.extract_all(archive, tempdir, szip_path)
                inf_path = os.path.join(tempdir, iia)

    if inf_path and is_match(open(inf_path, 'rb').read(), des_driver.name):
        logging.info('use inf_path: {}, driver name: {}'.format(repr(inf_path), repr(des_driver.name)))
        sysdrivers.add_by_inf(inf_path, des_driver.name)
        logging.info('done')
    else:
        logging.warning('bad driver -> name: {}, archive: {}, inf_path:{}'.format(repr(des_driver.name),
                                                                                  repr(des_driver.archive),
                                                                                  repr(des_driver.inf_path)))
        raise InstallationFailed


def install_printer(printer_name, driver_name, port_name):
    from wpi.printer import Printers
    sysprinters = Printers()

    if printer_name not in sysprinters.keys():
        printer = sysprinters.make_new(printer_name)
        printer.driver_name = driver_name
        printer.port_name = port_name
        sysprinters.save()
    else:
        logging.warning('printer: {} already existed, abord.'.format(repr(printer_name)))


def install(printers, drivers_dir):
    for p in printers:
        try:
            install_driver(p.driver, drivers_dir)
            install_port(p.port)
            install_printer(p.name, p.driver.name, p.port.name)

        except InstallationFailed:
            logging.error('install printer: {} failed.'.format(repr(p.name)))


def print_head():
    from shutil import get_terminal_size

    conlose_len = get_terminal_size()[0] - 2
    cl = conlose_len

    header_len = min([62, cl]) - 2
    hl = header_len

    item_left = 2

    def _(s):
        return s
        # return '{:^{}}'.format(s, cl)

    def sharp():
        s = '#' * hl
        return _(s)

    def _wpi():
        s = '{:^{}}'.format('Windows Printer Installer', hl)
        return _(s)

    def _item(k, v):
        s = '{}: {}'.format(k, v)
        s = (' ' * item_left) + s
        s += ' ' * (hl - len(s))
        return _(s)

    ss = [
        sharp(),
        '',
        _wpi(),
        '',
        _item('Version', version.__version__),
        _item('License', 'LGPL v3'),
        _item('Author', 'Chen Meng'),
        _item('HomePage', 'https://github.com/meng89/wpi'),
        sharp(),
    ]

    for _ in ss:
        print(_)


def main():
    from wpi.env import is_exe
    from wpi.log import set_file_handler, set_stream_handler

    if not ctypes.windll.shell32.IsUserAnAdmin():
        print('Not run as Administrator!')
        sys.exit()

    args = list()
    kwargs = dict()

    for _ in sys.argv[1:]:
        p_a = _.split('=', 1)
        if len(p_a) == 1:
            args.append(_)
        else:
            kwargs[p_a[0]] = p_a[1]

    log_filename = datetime.datetime.now().strftime('%Y_%m_%d-%H_%M_%S_%f') + '.log.txt'
    os.makedirs(LOGS_DIR, exist_ok=True)

    set_file_handler(os.path.join(LOGS_DIR, log_filename))
    set_stream_handler()

    log_sys_info()

    os.chdir(tempfile.gettempdir())

    if is_exe():
        exe_main(*args, **kwargs)
    else:
        script_main(*args, **kwargs)


def exe_main(ps=None, drivers=None):
    from wpi.env import exe_dir

    if drivers is not None:
        drivers_dir = drivers
    else:
        drivers_dir = os.path.join(exe_dir(), DEFAULT_DIRIVERS_NAME)

    if ps is not None:
        module_ = load_module(ps)
    elif os.path.exists(os.path.join(exe_dir(), DEFAULT_PS_NAME)):
        module_ = load_module(os.path.join(exe_dir(), DEFAULT_PS_NAME))
    else:
        module_ = None

    if module_ is not None:
        install(module_.printers, drivers_dir)
        sys.exit()

    interactive_loop(drivers_dir, exe_dir())


def script_main(ps=None, drivers=None):
    user_wpi_dir = os.path.join(os.getenv('LOCALAPPDATA'), 'wpi')

    if drivers is not None:
        drivers_dir = drivers
    else:
        drivers_dir = os.path.join(user_wpi_dir, DEFAULT_DIRIVERS_NAME)

    if ps is not None:
        module_ = load_module(ps)
    else:
        module_ = None

    if module_ is not None:
        install(module_.printers, drivers_dir)
        sys.exit()

    interactive_loop(drivers_dir, user_wpi_dir)


def interactive_loop(drivers_dir, m_target_dir):
    logging.info('m command target dir: '.format(m_target_dir))

    print_head()
    while True:
        print()
        print('Please input a command or printers file:\n' +
              '  m  Make sample of "ps" and make drivers structure directories.\n' +
              '  l  List all driver names in an archive or an inf file.\n' +
              '  q  Quit.')
        print('Cmd or "ps" file: ', end='')

        user_input = input().strip()

        if user_input.lower() == 'm':
            copy_text_file(original_ps_sample_path(), os.path.join(m_target_dir, USER_SAMPLE_PS_NAME), even_exists=True)
            target_drivers_dir = os.path.join(m_target_dir, DEFAULT_DIRIVERS_NAME)
            make_driver_dir_structure(target_drivers_dir)

        elif user_input.lower() == 'l':
            list_driver_loop()

        elif user_input.lower() in ('q', 'quit', 'e', 'exit'):
            break

        elif user_input.strip().lower().endswith('.py'):
            module_ = load_module(user_input.strip())
            install(module_.printers, drivers_dir)


def list_driver_loop():
    from wpi.archivelib import read
    from wpi.env import get_szip_path

    def _list(models, indent=4, step=4):
        for platform, _ in models.items():
            print('{}{}, {}:'.format(' '*indent, repr(platform[1]), repr(platform[2])))
            for key in _.keys():
                print('{}{}'.format(' '*indent + ' '*step, repr(key)))

    while True:
        print()
        print('Archive or inf(r to return): ', end='')
        user_input = input().strip()

        if user_input.lower() in ('r', 'q', 'e'):
            return

        elif user_input.lower().endswith('.inf'):
            _models = _get_models(open(user_input, 'rb').read())
            _list(_models)

        else:
            for inf in list_infs(user_input):
                _models = _get_models(read(user_input, inf, get_szip_path()))
                if _models:
                    print()
                    print('  {}:'.format(repr(inf)))
                    _list(_models, indent=6)


def make_driver_dir_structure(drivers_dir):
    for bit in ALL_BITS:
        for os_ in ALL_OS:
            try:
                os.makedirs(os.path.realpath(os.path.join(drivers_dir, bit, os_)), exist_ok=True)
            except OSError:
                pass


def copy_text_file(source, target, even_exists=False, lf2crlf=True):
    import io

    if lf2crlf:
        with io.StringIO(newline='\r\n') as s_strio:
            s_strio.writelines(open(source, 'r', encoding='utf8').readlines())
            s_strio.seek(0)
            s_bytes = bytes(s_strio.read(), encoding='utf8')
    else:
        s_bytes = open(source, 'rb').read()

    def _write_target():
        os.makedirs(os.path.split(target)[0], exist_ok=True)
        with open(target, 'wb') as f:
            f.write(s_bytes)

    if os.path.exists(target):
        if even_exists:
            if s_bytes != open(target, 'rb').read():
                _write_target()
        else:
            pass
    else:
        _write_target()


def original_ps_sample_path():
    from wpi import ps_sample
    from wpi.env import is_exe, meipass_path, BUNDLE_DATA_FOLDER

    if is_exe():
        return os.path.join(meipass_path(), BUNDLE_DATA_FOLDER,
                            os.path.splitext(os.path.split(ps_sample.__file__)[1])[0] + '.py')
    else:
        return ps_sample.__file__


def log_sys_info():
    logging.info('OS bit: {}'.format(CUR_BIT))
    logging.info('OS release: {}'.format(CUR_OS))
    logging.info('Python bit: {}'.format(PYTHON_BIT))
    logging.info('Python sys.version: {}'.format(sys.version))


if __name__ == '__main__':
    main()
